import { ConsoleMixin } from 'ferda/ConsoleMixin'
import { DatetimeMixin } from 'ferda/DatetimeMixin'
import { HttpMixin } from 'ferda/HttpMixin'
import { NotificationsMixin } from 'ferda/NotificationsMixin'
import { OBJECT_TYPE } from 'ferda/RegistryMixin'
import { URLS as FERDA_URLS } from 'ferda/Urls'
import { mapGetters } from 'vuex'

import { URLS } from 'ferda_verification/Urls'

export const VERIFICATION_STATUS = Object.freeze({
    ACCESSED: 'accessed',
    CANCELED: 'canceled',
    CONSTRAINED: 'constrained',
    CREATED: 'created',
    FAILED: 'failed',
    FUTILE: 'futile',
    INITIALIZED: 'initialized',
    MANUAL: 'manual',
    PAUSED: 'paused',
    PENDING: 'pending',
    SUCCEEDED: 'succeeded',
})

export const VERIFICATION_COLOR = Object.freeze({
    [VERIFICATION_STATUS.ACCESSED]: '#98c6ea',
    [VERIFICATION_STATUS.CANCELED]: '#98c6ea',
    [VERIFICATION_STATUS.CONSTRAINED]: '#fa93ab',
    [VERIFICATION_STATUS.CREATED]: '#eeeeee',
    [VERIFICATION_STATUS.FAILED]: '#fa93ab',
    [VERIFICATION_STATUS.FUTILE]: '#fa93ab',
    [VERIFICATION_STATUS.INITIALIZED]: '#eeeeee',
    [VERIFICATION_STATUS.MANUAL]: '#dca3f8',
    [VERIFICATION_STATUS.PAUSED]: '#dca3f8',
    [VERIFICATION_STATUS.PENDING]: '#98c6ea',
    [VERIFICATION_STATUS.SUCCEEDED]: '#65e0b8',
})

export const VERIFICATION_LIST_PARAMS = Object.freeze({
    DOMAIN_ID: 'domain_id',
    STATUSES: 'statuses',
    MANUAL_CHECK_REQUEST: 'manual_check_request',
})

export const VERIFICATION_EVENT_TYPE = Object.freeze({
    CANCELED: 'canceled',
    CREATE: 'create',
    CHECK_DOMAINS: 'check_domains',
    COMPLETE: 'complete',
    DELAY_STATUS: 'delay_status',
    FAILED: 'failed',
    FUTILE: 'futile',
    MANUAL: 'manual',
    OUTZONE: 'outzone',
    RESOLVE_MANUAL_CHECK: 'resolve_manual_check',
    RESUME: 'resume',
    REQUEST_MANUAL_CHECK: 'request_manual_check',
    SEND_MESSAGE: 'send_message',
    SET_CONTACT_REPRESENTATIVE: 'set_contact_representative',
    USE_METHOD: 'use_method',
    VERIFY: 'verify_contact_data',
    VERIFY_TOKEN: 'verify_token',
})

export const VERIFICATION_ACTION = Object.freeze({
    ADD: 'add_to_verifications',
    APPROVE_AUTOMATIC: 'approve_automatic',
    APPROVE_FUTILE: 'approve_futile',
    APPROVE_MANUAL: 'approve_manual',
    CANCEL: 'cancel',
    CONFIRM: 'confirm',
    DELAY_STATUS: 'delay_status',
    REJECT: 'reject',
    RESEND_EMAIL: 'resend_email',
})

export const VERIFICATION_MESSAGE_TYPE = Object.freeze({ VERIFICATION_INITIAL: 'verification_initial' })

export const VERIFICATION_DELAY_DATE_TYPE = Object.freeze({ FUTILE: 'futile', OUTZONE: 'outzone', RETURN: 'return' })

export const VerificationMixin = {
    mixins: [ConsoleMixin, DatetimeMixin, HttpMixin, NotificationsMixin],
    data() {
        return {
            is_table_opened: false,
        }
    },
    computed: {
        ...mapGetters('general', ['permissions']),
    },
    created() {
        this.PAGE_NUMBERS = Object.freeze([10, 20, 50, -1])
        this.OBJECT_TYPE = OBJECT_TYPE
    },
    methods: {
        /**
         * Toggles list section
         */
        async toggle_list(reload_callback, is_refresh = false) {
            this.is_table_opened = is_refresh ? true : !this.is_table_opened

            if (this.is_table_opened) {
                await reload_callback()
            }
        },
        /**
         * Gets verification list
         *
         * @param { Object } params - URLSearchParams query
         * @returns { Object } - null in case of server error
         * @returns { Object } - response with data and pagination
         */
        async get_verification_list_data(params) {
            let response
            try {
                response = await this.http_get(URLS.verification_list_ajax(), { params })
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_SOMETHING_WENT_WRONG)
                return null
            }
            return response.data
        },
        /**
         * Gets verification action form title
         *
         * @param { String } last_action_type - last clicked action
         * @returns { String } - desired title
         */
        get_verification_form_title(last_action_type) {
            switch (last_action_type) {
            case VERIFICATION_ACTION.ADD:
                return this.$t('ferda_verification_common:add_to_verifications')
            case VERIFICATION_ACTION.APPROVE_AUTOMATIC:
                return this.$t('ferda_verification_common:approve_automatic')
            case VERIFICATION_ACTION.APPROVE_FUTILE:
                return this.$t('ferda_verification_common:approve_futile')
            case VERIFICATION_ACTION.APPROVE_MANUAL:
                return this.$t('ferda_verification_common:approve_manual')
            case VERIFICATION_ACTION.CANCEL:
                return this.$t('ferda_verification_common:cancel_verification')
            case VERIFICATION_ACTION.CONFIRM:
                return this.$t('ferda_verification_common:confirm_verification')
            case VERIFICATION_ACTION.DELAY_STATUS:
                return this.$t('ferda_verification_common:delay_status')
            case VERIFICATION_ACTION.REJECT:
                return this.$t('ferda_verification_common:reject_verification')
            case VERIFICATION_ACTION.RESEND_EMAIL:
                return this.$t('ferda_verification_common:resend_email')
            default:
                return ''
            }
        },
        /**
         * Gets verification list params
         *
         * @param { Array<String> } statuses - wanted statuses
         * @param { String } domain_id - domain ID
         * @param { Boolean } manual_check_request
         * @returns { Object } - URLSearchParams query
         */
        get_verification_list_params(statuses, domain_id = null, manual_check_request = null) {
            const params = new URLSearchParams()
            for (const status of statuses) {
                params.append(VERIFICATION_LIST_PARAMS.STATUSES, status)
            }

            if (domain_id) {
                params.append(VERIFICATION_LIST_PARAMS.DOMAIN_ID, domain_id)
            }

            if (manual_check_request) {
                params.append(VERIFICATION_LIST_PARAMS.MANUAL_CHECK_REQUEST, manual_check_request)
            }
            return params
        },
        /**
         * Gets parameters for verification modification
         *
         * @param { String } last_action_type - desired action type
         * @param { Any } payload - any additional needed info
         * @returns { Object } - URLSearchParams query
         */
        get_verification_modify_params(last_action_type, payload = null) {
            const params = new URLSearchParams()
            switch (last_action_type) {
            case VERIFICATION_ACTION.RESEND_EMAIL:
                params.append('message_type', VERIFICATION_MESSAGE_TYPE.VERIFICATION_INITIAL)
                return params
            case VERIFICATION_ACTION.DELAY_STATUS:
                params.append(VERIFICATION_DELAY_DATE_TYPE.FUTILE, payload.futile)
                params.append(VERIFICATION_DELAY_DATE_TYPE.OUTZONE, payload.outzone)
                return params
            }
            return params
        },
        /**
         * Adds contact to verifications
         *
         * @param { String } contact_id - contact ID
         * @returns { Object } - null in case of server error
         */
        async add_to_verifications(contact_id, context) {
            const params = new URLSearchParams()
            params.append('contact_id', contact_id)
            params.append('reason', context.reason)
            if (context.url) {
                params.append('reason_url', context.url)
            }

            try {
                await this.http_post(URLS.add_to_verifications_ajax(), {}, {
                    params,
                    headers: { 'X-CSRFToken': this.$cookies.get('csrftoken') },
                })
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_SOMETHING_WENT_WRONG)
                return null
            }
        },
        /**
         * Approves domain deletion
         *
         * @param { String } verification_id - verification ID
         * @returns { Object } - null in case of server error
         */
        async approve_futile(verification_id) {
            try {
                await this.http_put(URLS.approve_futile_ajax(verification_id), {}, {
                    headers: { 'X-CSRFToken': this.$cookies.get('csrftoken') },
                })
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_SOMETHING_WENT_WRONG)
                return null
            }
        },
        /**
         * Cancels verification
         *
         * @param { String } verification_id - verification ID
         * @returns { Object } - null in case of server error
         */
        async cancel_verification(verification_id) {
            try {
                await this.http_put(URLS.cancel_verification_ajax(verification_id), {}, {
                    headers: { 'X-CSRFToken': this.$cookies.get('csrftoken') },
                })
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_SOMETHING_WENT_WRONG)
                return null
            }
        },
        /**
         * Gets verification by ID
         *
         * @param { String } verification_id - verification ID
         * @param { Object } abort_controller
         * @returns { Object } - null in case of server error
         * @returns { Object } - particular verification
         */
        async get_verification(verification_id, abort_controller = null) {
            let response
            try {
                response = await this.http_get(URLS.verification_detail_ajax(verification_id),
                    { signal: abort_controller?.signal || null })
            } catch (error) {
                const error_message = this.is_http_cancel(error) ?
                    this.INFO_REQUEST_ABORTED : this.ERROR_SOMETHING_WENT_WRONG
                this.log_error(error)
                this.notify_message(error_message)
                return null
            }
            return response.data
        },
        /**
         * Gets verification list by contact ID
         *
         * @param { String } contact_id - contact ID
         * @returns { Object } - null in case of server error
         * @returns { Object } - response with data and pagination
         */
        async get_verification_list_by_contact_id(contact_id) {
            let response
            try {
                response = await this.http_get(URLS.verification_list_ajax(), {
                    params: new URLSearchParams(`?contact_id=${contact_id}`),
                })
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_SOMETHING_WENT_WRONG)
                return null
            }
            return response.data
        },
        /**
         * Resends e-mail
         *
         * @param { String } verification_id - verification ID
         * @param { Object } params - URLSearchParams query
         * @returns { Object } - null in case of server error
         */
        async resend_email(verification_id, params) {
            try {
                await this.http_post(URLS.resend_email_ajax(verification_id), null, {
                    params,
                    headers: { 'X-CSRFToken': this.$cookies.get('csrftoken') },
                })
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_SOMETHING_WENT_WRONG)
                return null
            }
        },
        /**
         * Approves manual processing
         *
         * @param { String } verification_id - Verification ID
         * @returns { Object } - updated verification
         * @returns { Object } - null in case of server error
         */
        async approve_manual(verification_id) {
            let response
            try {
                response = await this.http_put(URLS.approve_manual_ajax(verification_id), null, {
                    headers: { 'X-CSRFToken': this.$cookies.get('csrftoken') },
                })
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_SOMETHING_WENT_WRONG)
                return null
            }
            return response.data
        },
        /**
         * Makes requests to delay status (either only futile or both futile and constrained)
         *
         * @param { String } verification_id - verification ID
         * @param { Object } params - URLSearchParams query
         * @returns { Object } - null in case of server error
         */
        async delay_status(verification_id, params) {
            const futile = parseInt(params.get(VERIFICATION_DELAY_DATE_TYPE.FUTILE))
            const outzone = parseInt(params.get(VERIFICATION_DELAY_DATE_TYPE.OUTZONE))
            const headers = { 'X-CSRFToken': this.$cookies.get('csrftoken') }

            // Prepare futile at first as this is called always
            const params_futile = new URLSearchParams()
            params_futile.append('status', VERIFICATION_STATUS.FUTILE)
            params_futile.append('delay', futile)

            const actions = [this.http_put(URLS.delay_status_ajax(verification_id), null, {
                params: params_futile, headers,
            })]

            // In case of non-zero outzone move, append it as another Promise
            if (outzone !== 0) {
                const params_constrained = new URLSearchParams()
                params_constrained.append('status', VERIFICATION_STATUS.CONSTRAINED)
                params_constrained.append('delay', outzone)

                actions.push(
                    this.http_put(URLS.delay_status_ajax(verification_id), null, {
                        params: params_constrained, headers,
                    }),
                )
            }

            try {
                await Promise.all(actions)
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_SOMETHING_WENT_WRONG)
                return null
            }
        },
        /**
         * Makes particular action above verification
         *
         * @param { String } last_action_type - last action type
         * @param { Object } params - URLSearchParams query
         * @param { String } verification_id - verification ID
         * @param { String } contact_id - contact ID
         * @param { Array } confirm_data - data form manual check verification
         * @returns { Object } - info after success operation
         * @returns { Object } - null in case of server error or unknown action type
         */
        async make_verification_action(
            last_action_type,
            params,
            verification_id,
            contact_id = null,
            confirm_data = null,
            context = null,
        ) {
            switch (last_action_type) {
            case VERIFICATION_ACTION.ADD:
                return await this.add_to_verifications(contact_id, context)
            case VERIFICATION_ACTION.APPROVE_AUTOMATIC:
                return await this.approve_automatic(verification_id)
            case VERIFICATION_ACTION.APPROVE_FUTILE:
                return await this.approve_futile(verification_id)
            case VERIFICATION_ACTION.APPROVE_MANUAL:
                return await this.approve_manual(verification_id)
            case VERIFICATION_ACTION.CANCEL:
                return await this.cancel_verification(verification_id)
            case VERIFICATION_ACTION.CONFIRM:
                return await this.manual_check_verification(true, verification_id, confirm_data)
            case VERIFICATION_ACTION.DELAY_STATUS:
                return await this.delay_status(verification_id, params)
            case VERIFICATION_ACTION.REJECT:
                return await this.manual_check_verification(false, verification_id, confirm_data)
            case VERIFICATION_ACTION.RESEND_EMAIL:
                return await this.resend_email(verification_id, params)
            default:
                return null
            }
        },
        /**
         * Gets specified date for particular verification
         *
         * @param { Object } - object with specified events and next events
         * @param { String } desired_event - event to be searched
         * @param { Boolean } multiple - whether there can be more events
         * @returns { Array<Object> } - in case there are more events
         * @returns { Object } - null in case of missing event
         * @returns { Object } - particular event (latest according to datetime)
         * @returns { Object } - order event from newest to oldest
         */
        get_verification_event({ events, next_events }, desired_event = null, multiple = false) {
            if (multiple) {
                if (desired_event) {
                    // In case of multipe possible events, list all of them from events and next_events
                    return [...events, ...next_events].filter(event => event.type === desired_event)
                        .sort((a, b) => new Date(b.event_datetime) - new Date(a.event_datetime))
                }
                return [...events, ...next_events]
                    .sort((a, b) => new Date(b.event_datetime) - new Date(a.event_datetime))
            }

            // In case of single event, get the latest one, preferably from next_events and then from events
            const next_event = this.get_latest_event(next_events.filter(event => event.type === desired_event))
            if (next_event) {
                return next_event
            }

            const event = this.get_latest_event(events.filter(event => event.type === desired_event))
            if (event) {
                return event
            }
            return null
        },
        /**
         * Gets latest event according to datetime
         *
         * @param { Array<Object> } events - array of desired events
         * @returns { Object } - null in case of server error
         * @returns { Object } - desired event
         */
        get_latest_event(events) {
            if (!events.length) {
                return null
            }
            return events.reduce((a, b) => new Date(a.event_datetime) > new Date(b.event_datetime) ? a : b)
        },
        /**
         * Checks whether user can resume verification to automatic processing
         *
         * @param { String } status - verification status
         * @returns { Boolean } - can or can't resume
         */
        can_approve_automatic(status) {
            return [
                VERIFICATION_STATUS.PAUSED,
            ].includes(status) && this.permissions.includes('ferda_verification.can_resume')
        },
        /**
         * Checks whether user can cancel verification
         *
         * @param { String } status - verification status
         * @returns { Boolean } - can or can't cancel
         */
        can_cancel(status) {
            return ![
                VERIFICATION_STATUS.CANCELED,
                VERIFICATION_STATUS.MANUAL,
            ].includes(status) && this.permissions.includes('ferda_verification.can_cancel')
        },
        /**
         * Checks whether user can add contact to verifications
         * Allowed are only contacts not already included in verifications and not deleted
         *
         * @param { Object } contact - info about contact itself
         * @returns { Boolean } - can or can't create
         */
        can_create(contact) {
            const has_permission = this.permissions.includes('ferda_verification.can_create')
            const is_deleted = !!contact.object_info.events.unregistered?.timestamp

            return has_permission && !is_deleted
        },
        /**
         * Checks whether user can resend e-mail
         *
         * @param { String } status - verification status
         * @returns { Boolean } - can or can't resend
         */
        can_resend(status) {
            return [
                VERIFICATION_STATUS.ACCESSED,
                VERIFICATION_STATUS.CONSTRAINED,
                VERIFICATION_STATUS.FUTILE,
                VERIFICATION_STATUS.INITIALIZED,
                VERIFICATION_STATUS.PENDING,
            ].includes(status) && this.permissions.includes('ferda_verification.can_resend_message')
        },
        /**
         * Checks whether user can delay status
         *
         * @param { String } status - verification status
         * @returns { Boolean } - can or can't delay
         */
        can_delay(status) {
            return ![
                VERIFICATION_STATUS.MANUAL,
                VERIFICATION_STATUS.PAUSED,
            ].includes(status) && this.permissions.includes('ferda_verification.can_delay_status')
        },
        /**
         * Checks whether user can approve futile
         *
         * @returns { Boolean } - can or can't approve futile
         */
        can_approve_futile() {
            return this.permissions.includes('ferda_verification.can_approve_futile')
        },
        /**
         * Checks whether user can approve manual
         *
         * @param { String } status - verification status
         * @returns { Boolean } - can or can't approve manual
         */
        can_approve_manual(status) {
            return [
                VERIFICATION_STATUS.ACCESSED,
                VERIFICATION_STATUS.CONSTRAINED,
                VERIFICATION_STATUS.CREATED,
                VERIFICATION_STATUS.INITIALIZED,
                VERIFICATION_STATUS.PAUSED,
                VERIFICATION_STATUS.PENDING,
            ].includes(status) && this.permissions.includes('ferda_verification.can_approve_manual')
        },
        /**
         * Checks whether user can resolve manual check verification
         *
         * @returns { Bolean } - can or can't resolve manual check
         */
        can_resolve() {
            return this.permissions.includes('ferda_verification.can_resolve_manual_check')
        },
        /**
         * Sorts table columns
         *
         * @param { Array<Object> } items - table data
         * @param { Array<String> } sort_by - array of columns table is sorted by (always just 1)
         * @param { Array<Boolean> } is_desc - array of columns direction sort (always just 1)
         * @returns { Array<Object> } - sorted table data
         */
        sort_table(items, sort_by, is_desc) {
            return items.sort((first, second) => {
                if (sort_by[0] === 'create_datetime' || sort_by[0] === 'close_datetime') {
                    const first_date = new Date(first[sort_by])
                    const second_date = new Date(second[sort_by])
                    return is_desc[0] ? second_date - first_date : first_date - second_date
                }

                if (sort_by[0] === 'events') {
                    const first_date = new Date(this.get_verification_event(
                        first, VERIFICATION_EVENT_TYPE.OUTZONE,
                    ).event_datetime)

                    const second_date = new Date(this.get_verification_event(
                        second, VERIFICATION_EVENT_TYPE.OUTZONE,
                    ).event_datetime)
                    return is_desc[0] ?  second_date - first_date : first_date - second_date
                }

                if (sort_by[0] === 'manual_check_request') {
                    return is_desc[0] ? first[sort_by] - second[sort_by] : second[sort_by] - first[sort_by]
                }

                if (typeof first[sort_by] !== 'undefined') {
                    const first_val = first[sort_by].toLowerCase().trim()
                    const second_val = second[sort_by].toLowerCase().trim()
                    return is_desc[0] ? second_val.localeCompare(first_val) : first_val.localeCompare(second_val)
                }
                return items
            })
        },
        /**
         * Gets additional contact info
         *
         * @returns { Object } - null in case of server error
         * @returns { Object } - detail of contact
         */
        async get_contact_detail(contact_id) {
            let response
            try {
                response = await this.http_get(FERDA_URLS.object_info_ajax(OBJECT_TYPE.CONTACT, contact_id))
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_SOMETHING_WENT_WRONG)
                return null
            }
            return response.data
        },
        /**
         * Approves automatic processing
         *
         * @param { String } verification_id - verification ID
         * @returns { Object } - null in case of server error
         */
        async approve_automatic(verification_id) {
            try {
                await this.http_put(URLS.approve_automatic_ajax(verification_id), null,
                    { headers: { 'X-CSRFToken': this.$cookies.get('csrftoken') }},
                )
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_SOMETHING_WENT_WRONG)
                return null
            }
        },
        /**
         * Manually confirm or reject verification
         *
         * @param { Boolean } verified - confirmation or rejection of data
         * @param { String } verification_id
         * @param { Object } confirm_data
         * @returns { Object } - null in case of server error
        */
        async manual_check_verification(verified, verification_id, confirm_data) {
            const data = {
                verified: verified,
                contact_data: confirm_data.data,
                contact_history_id: confirm_data.contact_history_id,
            }
            try {
                await this.http_put(URLS.confirm_verification_ajax(verification_id),
                    data, { headers: { 'X-CSRFToken': this.$cookies.get('csrftoken') }},
                )
            } catch (error) {
                this.log_error(error)
                this.notify_message(this.ERROR_SOMETHING_WENT_WRONG)
                return null
            }
        },
    },
}
